#!/usr/bin/perl 
# $Header: /home/hiro/projects/dvr2/scripts/RCS/dvr2-common.pl,v 1.16 2012/12/18 17:16:27 hiro Exp $
#===============================================================================
#
#         FILE:  dvr2-common.pl
#
#        USAGE:  require "$mydir/dvr2-commnon.pl"
#
#  DESCRIPTION:  Common executive code
#
#       AUTHOR:  Copyright (c) 2011 Hiro Sugawara
#      VERSION:  1.0
#      CREATED:  10/24/11 10:25:23
#     REVISION:  ---
#===============================================================================

use strict;
use warnings;
use POSIX;
use DBI();
use File::Basename;
use XML::Simple;

our $myname;
our $mydir;
our $DVR2ROOT = '/var/multimedia/video';
my $config;
my $rsvTBL;
my $prgTBL;
my $chlTBL;
my $tsDIR;

sub parseXML {
	my $xmlfile = "$mydir/../settings/config.xml";
	my $xml = XML::Simple->new();
	$config = $xml->XMLin($xmlfile);

	$rsvTBL	= $config->{tbl_prefix} . 'reserveTbl';
	$prgTBL	= $config->{tbl_prefix} . 'programTbl';
	$chlTBL	= $config->{tbl_prefix} . 'channelTbl';
	$tsDIR	= basename $config->{spool};
}

my $dbh;
sub openDB {
	if (defined($dbh)) {
		return;
	}
	parseXML();
	$dbh = DBI->connect('DBI:mysql:database=' . $config->{db_name} .
				';host=' . $config->{db_host},
				$config->{db_user}, $config->{db_pass}, {'RaiseError' => 1});
	die "$myname: can't open database"	unless defined($dbh);
	$dbh->do('set names UTF8');
}

sub closeDB {
	$dbh->disconnect();
	undef $dbh;
}

#
# Read EPG from DB
#	DB record is returned in $rec hash reference.
#
sub getEPG {
	my ($id, $rec) = @_;
	openDB();
	my @tables = ();

	if ($id =~ /^\d+$/) {
		push @tables, sprintf("$rsvTBL WHERE program_id='%s'", $id);
		push @tables, sprintf("$prgTBL WHERE id='%s'", $id);
	} elsif ($id =~ /\.ts$/) {
		my $tsfile = basename $id;
		push @tables, sprintf("$rsvTBL WHERE path='%s'", $tsfile);
	} else {
		die "$myname: bad tsORpid ($id)\n";
	}

	my $query;
	my $sth;
	my $row;

	foreach my $table (@tables) {
		$query = sprintf "SELECT * FROM %s", $table;
		$sth = $dbh->prepare($query);
		$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
		$row = $sth->fetchrow_hashref();
		defined($row) && last;
		$sth->finish();
	}
	return undef unless (defined($row));
	%$rec = %$row;
	$sth->finish();
	if (!defined($rec->{program_id})) {
		$rec->{program_id} = $row->{id};
	}

	$query = sprintf "SELECT * FROM $prgTBL WHERE id='%s'", $rec->{program_id};
	$sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	$row = $sth->fetchrow_hashref();
	$rec->{starttime} = $row->{starttime};
	$rec->{endtime} = $row->{endtime};
	$rec->{program_disc} = $row->{program_disc};
	$sth->finish();

	$query = sprintf "SELECT * FROM $chlTBL WHERE channel_disc='%s'",
				$rec->{channel_disc};
	$sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	$row = $sth->fetchrow_hashref();
	$rec->{station} = $row->{name};
	$sth->finish();
}

#
# This is stupid. The installed version of iconv does not have
# set_attr/iconvctl to tolerate bad characters. Use shell.
#
sub iconv_utf2sj {
	my ($str) = @_;
	my $conv = `echo '$str'|iconv -c -s -f utf-8 -t shift_jis//TRANSLIT`;
	chomp $conv;
	return $conv;
}

#
# Look for an existing IEPG file having the TS file name
#
sub searchIEPG {
	my ($id, $iepgref)	= @_;
	$id = basename $id;
	$id =~ s/\./\\./;
	my $iepgFile = `grep "^tsfile: $id" $DVR2ROOT/recorded/*.iepg`;
	return undef	unless ($iepgFile);
	chomp $iepgFile;
	$iepgFile =~ s/:.*$//;

	open IEPG, $iepgFile	|| return undef;
	while (<IEPG>) {
		chomp;
		last unless ($_);
		my ($key, $val) = split /: /;
		$iepgref->{$key} = $val;
	}
	close IEPG;
	return basename $iepgFile, '.iepg';
}

#
# Read an EPG record and convert it into an iEPG in given hash reference
#	Returns file name core string
#
sub getIEPG {
	my ($id, $iepgref) = @_;
	my %rec;

	my $fname = searchIEPG($id, $iepgref);
	return $fname if (defined($fname));

	unless (defined(getEPG($id, \%rec))) {
		return undef;
	}
	my $channel	= $rec{'channel'};
	my $type	= $rec{'type'};
	my $starttime	= $rec{'starttime'};
	my $endtime	= $rec{'endtime'};
	my $program_id	= $rec{'program_id'};
	my $path	= $rec{'path'};
	my $program_disc	= $rec{'program_disc'};
	my $job		= $rec{'job'};

	return undef unless(defined($starttime));

	my ($year, $month, $date)	= split /[\- :]/, $starttime;
	my $start	= substr((split / /, $starttime)[1], 0, 5);
	my $end		= substr((split / /, $endtime)[1], 0, 5);

	my $program_title	= iconv_utf2sj($rec{title});
	# Remove [x] icons
	$program_title	=~ s/\x{81}\x{79}..\x{81}\x{7a}//g;
	my $program_subtitle	= iconv_utf2sj($rec{description});
	my $station	= iconv_utf2sj($rec{station});

	if (defined($iepgref)) {
		%$iepgref = (
		'Content-type'	=> 'application/x-tv-program-info; charset=shift_jis',
		'version'	=> '1',
		'station'	=> $station,
		'year'	=> $year,
		'month'	=> $month,
		'date'	=> $date,
		'start'	=> $start,
		'end'	=> $end,
		'program-title'	=> $program_title,
		'program-subtitle' => $program_subtitle,
		'd-channel'	=> "$type$channel",
		'program_id'	=> $program_id,
		'tsfile'	=> $path,
		'tsdir'		=> "$DVR2ROOT/$tsDIR",
		'program_disc'	=> $program_disc,
		'at-job'	=> $job,
		);
	}
	my @start = split /:/, $start;
	my @end   = split /:/, $end;
	my $hour = $end[0] - $start[0];
	my $min  = $end[1] - $start[1];
	if ($min < 0) {
		$min += 60;
		$hour--;
	}
	if ($hour < 0) {
		$hour += 24;
	}

	return sprintf "%04d%02d%02d%02d%02d-9%02d-%02d%02d-%08x",
					$year, $month, $date,
					substr($start, 0, 2), substr($start, 3, 2),
					substr($channel, -2), $hour, $min, $program_id;
}

#
# Get full path name
#
sub getFname {
	my ($id, $tab, $suffix, $iepgref)	= @_;
	my %iepg;
	$iepgref = \%iepg unless defined($iepgref);
	my $fname = getIEPG($id, $iepgref);
	return undef unless (defined($fname));
	if (defined($tab)) {
		$fname = sprintf "%s/%s/%s", $DVR2ROOT, $tab, $fname;
	}
	if (defined($suffix)) {
		$fname .= ".$suffix";
	}
	return $fname;
}

#
# Create a new iEPG with file name and iEPG
#
sub createIEPG2 {
	my ($fname, $iepgref) = @_;

	open IEPG, ">$fname"	|| die "$myname: can't open $fname\n";
	foreach my $iepg (keys(%$iepgref)) {
		my $val = $iepgref->{$iepg};
		unless (defined($val)) {
			printf STDERR "%s: UNDEFINED val for %s in %s\n",
					"createIEPG2", $iepg, $fname;
			$val = 'UNKNOWN';
		}
		printf IEPG "%s: %s\n", $iepg, $val;
	}
	close IEPG;
}

#
# Create a new iEPG
#
sub createIEPG {
	my ($id, $tab) = @_;
	my %iepg;

	my $fname = getFname($id, $tab, 'iepg', \%iepg);
	die "$myname: cannot get file name for $id (createIPEG)\n"
		unless (defined($fname));
	createIEPG2($fname, \%iepg);
}

#
# Append items to an existing iEPG
#
sub appendIEPG {
	my ($id, $tab, $iepgref, $suffix) = @_;
	$suffix = 'iepg' unless (defined($suffix));
	my $fname = getFname($id, $tab, $suffix);
	die "$myname: cannot get file name for $id (appendIPEG)\n"
		unless (defined($fname));

	open IEPG, ">>$fname"	|| die "$myname: can't open $fname";
	foreach my $iepg (keys(%$iepgref)) {
		printf IEPG "%s: %s\n", $iepg, $iepgref->{$iepg};
	}
	close IEPG;
}

#
# Move an existing iEPG from dir to another
#
sub moveIEPG {
	my ($id, $from, $to, $append, $suffix2) = @_;
	$suffix2 = 'iepg' unless (defined($suffix2));
	my $fname = getFname($id, $from, 'iepg');
	die "$myname: cannot get file name for $id (moveIPEG)\n"
		unless (defined($fname));
	my $fname2 = getFname($id, $to, $suffix2);
	unless ( -e $fname ) {
		$fname =~ s/[[:xdigit:]]{8}\.iepg$/*.iepg/;
		$fname = `ls $fname 2> /dev/null | head -1`;
		($fname) || die "$myname: can't find $id";
	}
	rename $fname, $fname2 || die "$myname: can't rename $fname to $fname2";
	appendIEPG($id, $to, $append, $suffix2)	 if (defined($append));
}

sub moveIEPG2REC {
	my ($id, $append)	= @_;
	moveIEPG($id, 'scheduled', 'recorded', $append);
}

sub renameIEPG2SKIP {
	my ($id, $append)	= @_;
	moveIEPG($id, 'scheduled', 'scheduled', undef, 'skip');
}

#
# Stop AT job
#
sub stopAT {
	my ($job) = @_;
	if (defined($job)) {
		system "atrm $job 2> /dev/null";
	}
}

#
# Delete an existing EPG reservation
#
sub delReserve {
	my ($id) = @_;
	my %rec;
	die "$myname: cannot get file name for $id (delReserve)\n"
		unless (defined(getEPG($id, \%rec)));
	stopAT($rec{'job'});
	my $query = sprintf "DELETE FROM $rsvTBL WHERE program_id = '%s'",
					$rec{program_id};
	my $sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	$sth->finish();
}

#
# Delete an existing iEPG
#
sub deleteIEPG {
	my ($id, $tab) = @_;
	my %iepg;
	my $fname = getFname($id, $tab, 'iepg', \%iepg);
	if (defined($fname)) {
		stopAT($iepg{'at-job'});
		unlink $fname;
	}
}

#
# Delete skip file
#
sub deleteSKIP {
	my ($id) = @_;
	my $fname = getFname($id, 'scheduled', 'skip');
	if (defined($fname)) {
		unlink $fname;
	}
}

#
# Skip file exists
#
sub existsSKIP {
	my ($id) = @_;
	my $fname = getFname($id, 'scheduled', 'skip');
	return ($fname && (-e $fname)) ? 1 : 0;
}

#
# Delete all relating files
#
sub deleteFiles {
	my ($id, $tab, @suffices) = @_;
	my $fbase = getFname($id, $tab);
	if (defined($fbase)) {
		foreach my $suffix (@suffices) {
			unlink "$fbase.$suffix";
		}
	}
}

#
# Set stat file content
#
sub setStat {
	my ($id, $stat) = @_;
	my $fname = getFname($id, 'recorded', 'stat');
	die "$myname: cannot get file name for $id (setStat)\n"
		unless (defined($fname));

	open STAT, ">$fname"	||
		die "$myname: can't open $fname";
	printf STAT "%s\n", $stat;
	close STAT;
}

#
# Kill recorder process
#
sub killRecorder {
	my ($id) = @_;
}

#
# Look for un-iEPG'ed scheduled recordings in DB to create their iEPG files
#
sub sync2epgAdd {
	openDB();

	my $query =
	"SELECT program_id FROM $rsvTBL WHERE starttime > ADDTIME(NOW(),'0:0:10')";
	my $sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	while ((my $row = $sth->fetchrow_hashref())) {
		my $program_id = $row->{program_id};
		my %iepg;
		my $fname = getFname($program_id, 'scheduled', 'iepg', \%iepg);
		defined($fname) && ((-e $fname)  || createIEPG2($fname, \%iepg));
	}
	$sth->finish();
}

#
# Look for stale scheduled recordings in DB and remove their iEPG files
#
sub sync2epgRemove {
	openDB();

	my $dir = "$DVR2ROOT/scheduled";

	opendir DIR, $dir	|| die "$myname: can't open $dir\n";
	while ((my $file = readdir DIR)) {
		my %iepg;
		($file =~ /\.iepg$/)	|| next;
		open IEPG, "$dir/$file"	|| next;
		while (<IEPG>) {
			chomp;
			my ($item, $value)	= split /: /;
			$iepg{$item} = $value;
		}
		close IEPG;
		exists($iepg{program_id})	|| next;
		my $query = sprintf "SELECT * FROM $rsvTBL WHERE program_id='%s'",
						$iepg{program_id};
		my $sth = $dbh->prepare($query);
		$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
		defined($sth->fetchrow_hashref())	|| unlink $file;
		$sth->finish();
	}
	closedir DIR;
}

#
# Look for aged recording in DB and delete it
#
sub deleteOldRec {
	openDB();

	my $query = "DELETE FROM $rsvTBL WHERE endtime < SUBDATE(NOW(), 14)";
	my $sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	$sth->finish();
}

#
# Look for garbage files and delete them
#
sub find2rm {
	my ($dir, $days, $opts)	= @_;
	if (substr($dir, 0, 1) ne '/') {
		$dir = "$DVR2ROOT/$dir";
	}
	my $expire = strftime "%Y-%m-%d", localtime(time - 86400 * $days);
	my $find = sprintf "find %s %s \\! -newermt %s -exec rm -f {} \\;",
				$dir,
				defined($opts) ? $opts : '',
				$expire;
	system $find;
}

sub deleteGarbage {
	my $_48hago = strftime "%Y-%m-%d", localtime(time - 86400 * 2);
	my $_96hago = strftime "%Y-%m-%d", localtime(time - 86400 * 4);
	my $find;

	find2rm('spool', 2, '-size 0');
	find2rm('recorded', 2, '-size 0');
	find2rm('spool', 4);
	find2rm('recorded', 4, "-name '*.ts' -or -name '*.bak'");
}

#
# Return 0 if no recording is scheduled in next X minutes
#
sub pt1free {
	my ($min)	= @_;
	$min = 10 unless defined($min);
	my $addtime = sprintf "0:%d:0", $min;

	openDB();
	my $query = "SELECT COUNT(*) FROM $rsvTBL WHERE " .
		"starttime >= NOW() AND starttime < ADDTIME(NOW(),'$addtime')";

	my $retry = 10;
	while ($retry-- >= 0) {
		my $sth = $dbh->prepare($query);
		$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
		my $row = $sth->fetchrow_hashref();
		$sth->finish();
		if (!$row->{'COUNT(*)'}) {
			closeDB();
			return 0;
		}
	}
	closeDB();
	return 1;
}

#
# Remove *.iepg that dont have at job
#
sub sync2ATRemove {
	my $whoami	= `whoami`;
	chomp $whoami;
	my @atq	= `atq | awk '/$whoami\$/ { print \$1 }'`;
	chomp @atq;
	my %atq = ();
	foreach (@atq) {
		$atq{$_} = 1;
	}

	foreach my $iepg (`ls  $DVR2ROOT/scheduled/*.iepg 2> /dev/null`) {
		chomp $iepg;
		my $atjob = `awk '/^at-job/ { print \$2 }' $iepg`;
		chomp $atjob;
		unlink $iepg unless exists($atq{$atjob});
	}
}

#
# Show all tunable channels in DB in "ch:station" format
#
sub showChannels {
	openDB();
	my %chName;
	my $query = "SELECT channel,name FROM Recorder_channelTbl WHERE " .
		"channel <> '0'";
	my $sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	while ((my $row = $sth->fetchrow_hashref())) {
		my $channel = $row->{channel};
		my $name = $row->{name};
		$chName{$channel} = $name;
	}

	foreach (sort {$a <=> $b} keys(%chName)) {
		printf "%d:%s\n", $_, $chName{$_};
	}
	$sth->finish();
	closeDB();
}

#
# Show current program title and other information
#	showIEPG(27) for current
#	showIEPG(27, "12:34") for starting 12:34 in the nearest future
#
sub showIEPG {
	my ($ch, $stime)	= @_;

	openDB();
	my $query = "SELECT * FROM $prgTBL WHERE channel = $ch AND ";

	if (!defined($stime)) {
		$query .= "starttime <= NOW() and NOW() < endtime";
	} else {
		my $hhmm = strftime "%H:%M", localtime(time);
		$query .= strftime("starttime = '%Y-%m-%d $stime:00'",
							localtime(time + (($hhmm le $stime) ? 0 : 86400)));
	}

	my $sth = $dbh->prepare($query);
	$sth->execute()	|| die "$myname: can't execute: " . $sth->errstr();
	my $row = $sth->fetchrow_hashref();
	if (defined($row)) {
		my $starttime = $row->{starttime};
		my $endtime = $row->{endtime};
		my ($year, $month, $date)	= split /[\- :]/, $starttime;
		my $start	= substr((split / /, $starttime)[1], 0, 5);
		my $end		= substr((split / /, $endtime)[1], 0, 5);
		print <<_EOS;
Content-type: application/x-tv-program-info; charset=utf8
year: $year
month: $month
date: $date
start: $start
end: $end
program-title: $row->{title}
program-subtitle: $row->{description}
channel: $ch
_EOS
	}
	$sth->finish();
	closeDB();
}

# THIS MUST BE 1 FOR SUCCESSFUL INCLUSION
1;
